#!/bin/bash
#
# bash-validator:bin/bash-validator-add-to-ignore.sh
#
set -u
set -o pipefail

PROJECT_HOME="$( dirname "$( dirname "$( realpath "${BASH_SOURCE[0]}" )" )" )"
declare -r PROJECT_HOME

# shellcheck disable=SC1091
source "${PROJECT_HOME}/apis/bash-common-api/bash-common-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/bash-validator-config-api.sh" || exit $?

function usage {
  cat <<EOF
Usage

  $( basename "$0" ) --help
  $( basename "$0" ) [--normalize] [--workspace DIR] PATTERN1 [PATTERN2]...
  $( basename "$0" ) [--normalize] --user PATTERN1 [PATTERN2]...
  $( basename "$0" ) --normalize [--workspace DIR]
  $( basename "$0" ) --normalize [--user]

Options

  -h,       --help      Display help message and exit.
  -U,       --user      Apply configuration at USER level.
  -d DIR,   --workspace DIR
                        Define workspace directory
                        (if ommited use current directory)
  -N,       --normalize
                        Normalize configuration file

EOF
}

function normalize_config {
  local -r config_filename="${1}"
  local -r config_safe="${config_filename}.safe"

  echoinfo "Normalize: '${config_filename}'"
  LC_ALL=C sort --unique <"${config_filename}" >"${config_safe}" || return $?
  mv -f "${config_safe}" "${config_filename}" || return $?
}

function add_to_ignore_config_file {
  local -r config_file="${1}"
  shift

  (
    for pattern in "$@" ; do
      echo "${pattern}"
    done
  ) >>"${config_file}"
}

function add_to_ignore {
  local config_file=
  case "${param_configuration_level}" in
    'USER')
      config_file="${BASH_VALIDATOR_IGNORE_USER}"
      ;;

    'WORKSPACE')
      config_file="${param_workspace_dir}/${BASH_VALIDATOR_IGNORE_WORKSPACE}"
      ;;

    *)
      echofatal "Internal error: param_configuration_level='${param_configuration_level}'"
      return 1
  esac

  add_to_ignore_config_file "${config_file}" "$@" || return $?

  if [ "${param_normalize}" = 'true' ] ; then
    normalize_config "${config_file}" || return $?
  fi
}

function main {
  local -a param_files=()
  local param=
  local param_configuration_level=
  local param_normalize=
  local param_workspace_dir=

  while [ "$#" -gt 0 ] ; do
    param="$1"
    shift

    case "${param}" in
      '-h'|'--help')
        usage
        return 0
        ;;

      '-U'|'--user')
        if [ -n "${param_configuration_level}" ] ; then
          echofatal "Parameters --user and --workspace can't be define twice."
          return 1
        fi
        param_configuration_level='USER'
        ;;

      '-d'|'--workspace')
        if [ -n "${param_workspace_dir}" ] ; then
          echofatal "Parameters --user and --workspace can't be define twice."
          return 1
        fi
        param_workspace_dir="$( P "$@" )" && shift
        param_configuration_level='WORKSPACE'
        ;;

      '-N'|'--normalize')
        param_normalize='true';
        ;;

       *)
         param_files+=( "${param}" )
         ;;
    esac
  done

  if [ "${#param_files[@]}" -eq 0 ] && [ "${param_normalize}" != 'true' ]; then
    echofatal 'At least one file name is required.'
    usage >&2
    return 1
  fi

  if [ -n "${param_workspace_dir}" ] ; then
    if [ ! -d "${param_workspace_dir}" ] ; then
      echofatal "Can't find directory: '${param_workspace_dir}'"
      return 1
    fi
  else
    param_workspace_dir="$( pwd )"
    param_configuration_level='WORKSPACE'
  fi

  add_to_ignore "${param_files[@]}" || return $?
}

main "$@" || exit $?
