#!/bin/bash
#
# bash-validator:bin/bash-validator.sh
#
set -u
set -o pipefail

PROJECT_HOME="$( dirname "$( dirname "$( realpath "${BASH_SOURCE[0]}" )" )" )"
declare -r PROJECT_HOME

# shellcheck disable=SC1091
source "${PROJECT_HOME}/apis/bash-common-api/bash-common-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/bash-validator-config-api.sh" || exit $?

declare -g SHELLCHECK_ERROR='false'

function usage {
  cat <<EOF
Usage:

  $( basename "$0" ) [--help]
  $( basename "$0" ) --sh-only [-v|-vv|--verbose] [-d|--workspace DIR] [--ignore-default-config] [--ignore-user-config] [--ignore-workspace-config]
  $( basename "$0" ) [-v|-vv|--verbose] [-d|--workspace DIR] [--ignore-default-config] [--ignore-user-config] [--ignore-workspace-config]

  -h,       --help      Display help message and exit.
            --sh-only   Check only *.sh files
  -v,       --verbose   Enable verbose mode
            -vv         Enable very verbose mode
  -d DIR,   --workspace DIR
                        Define start directory and recurse in it.
                        (if ommited use current directory)
            --ignore-default-config
                        Ignore default configuration
            --ignore-user-config
                        Ignore USER configuration
            --ignore-workspace-config
                        Ignore WORKSPACE configuration


Some files can be globally ignored using '${BASH_VALIDATOR_IGNORE_USER}' (user configuration)
or from the folder to analyze using '${BASH_VALIDATOR_IGNORE_WORKSPACE}' (workspace configuration).

EOF
}

function echoinfo_verbose {
  if [ "${VERBOSE:-}" = "true" ]; then
    echoinfo "$@"
  fi
}

function echowarn_verbose {
  if [ "${VERBOSE:-}" = "true" ]; then
    echowarn "$@"
  fi
}

function do_shellcheck {
  local -r file="$1"
  local -r shell="${2:-}"

  local -a options=()

  if [ -n "${shell}" ] ; then
    options+=("--shell=${shell}")
  fi

  shellcheck "${options[@]}" "${file}" || SHELLCHECK_ERROR=true
}

function __get_config__ {
  if [ "${param_ignore_default_config}" != 'true' ] ; then
    cat <<EOF
*/.git/*
*/.metadata/*
*/.venv/*
EOF
  fi
  if [ "${param_ignore_user_config}" != 'true' ] ; then
    if [ -f "${BASH_VALIDATOR_IGNORE_USER}" ] ; then
      grep -v '^#' "${BASH_VALIDATOR_IGNORE_USER}" || true
    else
      echowarn_verbose "User configuration not found: '${BASH_VALIDATOR_IGNORE_USER}'"
    fi
  fi
  if [ "${param_ignore_workspace_config}" != 'true' ] ; then
    if [ -f "${BASH_VALIDATOR_IGNORE_WORKSPACE}" ] ; then
      grep -v '^#' "${BASH_VALIDATOR_IGNORE_WORKSPACE}" || true
    else
      echowarn_verbose "Workspace configuration not found: '${BASH_VALIDATOR_IGNORE_WORKSPACE}'"
    fi
  fi
}

function __load_ignore_config__ {
  __get_config__ | sort | uniq
}

function __build_not_path_entries__ {
  local first='true'
  local fixedLine=
  while read -r fixedLine ; do
    if [ -n "${fixedLine}" ] ; then
      if [ "${first}" = 'true' ] ; then
        first='false'
      else
        echo '-or'
      fi
      echo '-path'
      echo "${fixedLine}"
    fi
  done < <( __load_ignore_config__ | sed -e 's,\.,\\.,g' )
}

function build_not_path_from_config {
  local opts_paths=
  opts_paths="$( __build_not_path_entries__ )"

  if [ -n "${opts_paths}" ] ; then
    echo '-not'
    echo '('
    echo "${opts_paths}"
    echo ')'
  fi
}

function find_files {
  local -r shonly="$1"

  local -a find_opts=(-type f)

  local path_raw_opts=
  path_raw_opts="$( build_not_path_from_config )"

  if [ -n "${path_raw_opts}" ] ; then
    local opt=
    while read -r opt ; do
      if [ -n "${opt}" ] ; then
        find_opts+=("${opt}")
      fi
    done <<<"${path_raw_opts}"
  fi

  if [ "${shonly}" = 'true' ] ; then
    find_opts+=(-name "*.sh")
  fi
  find_opts+=(-print0 )

  (
    [ "${MESSAGE_API_SHOW_DEBUG:-}" = 'true' ] && set -x
    find "${path}" "${find_opts[@]}"
  ) | sort -z
}

function do_shellcheck_for_sh {
  local -r path="$1"

  local file=
  while IFS= read -r -d $'\0' file; do
    if [ -n "${file}" ] ; then
      echoinfo_verbose "${file}"
      do_shellcheck "${file}"
    fi
  done < <( find_files 'true' )
}

function do_shellcheck_for_all {
  local -r path="$1"

  local file=
  while IFS= read -r -d $'\0' file; do
    if [ -n "${file}" ] ; then
      do_shellcheck_for_all_check_entry "${file}" || return $?
    fi
  done < <( find_files "false" )
}

function do_shellcheck_for_all_check_entry {
  local -r file="$1"

  local __headline__=
  read -r -d '' __headline__ < <( head -c 100 "${file}" | head -n 1 )
  local -r __headline__

  local -r magic="${__headline__:0:2}"
  if [ "${magic}" = '#!' ] ; then
    local header_command=
    header_command="$( sed -E 's,..[[:blank:]]*([^[:blank:]]*)(.*)$,\1,g' <<<"${__headline__}" )" || return $?
    local -r header_command

    case "${header_command}" in
      '/bin/bash')
        echoinfo_verbose "${file}:bash"
        do_shellcheck "${file}" 'bash'
        ;;

      '/bin/sh')
        echoinfo_verbose "${file}:sh"
        do_shellcheck "${file}" 'sh'
        ;;

      '/bin/ksh')
        echoinfo_verbose "${file}:ksh"
        do_shellcheck "${file}" 'ksh'
        ;;

      '/usr/bin/env')
        local header_param=
        header_param="$( sed -E 's,..[[:blank:]]*([^[:blank:]]*)[[:blank:]]*([^[:blank:]]*)(.*)$,\2,g' <<<"${__headline__}" )" || return $?
        local -r header_param

        case "${header_param}" in
          'bash')
            echoinfo_verbose "${file}:bash (/usr/bin/env) '${header_param}'"
            do_shellcheck "${file}" 'bash'
            ;;

          'sh')
            echoinfo_verbose "${file}:sh (/usr/bin/env) '${header_param}'"
            do_shellcheck "${file}" 'sh'
            ;;

          'ksh')
            echoinfo_verbose "${file}:ksh (/usr/bin/env) '${header_param}'"
            do_shellcheck "${file}" 'ksh'
            ;;

          *)
            echodebug "${file} IGNORE '${__headline__}': (/usr/bin/env) '${header_param}'"
            ;;
        esac
        ;;

      *)
        if [ "${file}" = "${file%.sh}.sh" ] ; then
          echoinfo_verbose "${file}:end-with-sh"
          do_shellcheck "${file}"
        fi
        ;;
    esac
  fi
}

function main {
  local param=
  local param_function='all'
  local param_ignore_default_config=
  local param_ignore_user_config=
  local param_ignore_workspace_config=
  local param_workspace_dir='.'

  if [ -z "${MESSAGE_API_HIDE_DATE:-}" ] ; then
    declare -grx MESSAGE_API_HIDE_DATE='true'
  fi

  while [ "$#" -gt 0 ] ; do
    param="$1"
    shift

    case "${param}" in
      '-h'|'--help')
        usage
        return 0
        ;;

      '--sh-only')
        param_function='sh_only'
        ;;

      '-v'|'--verbose')
        VERBOSE='true'
        echoinfo_verbose 'Verbose mode enabled'
        ;;

      '-d'|'--workspace')
        param_workspace_dir="$( P "$@" )" && shift
        ;;

      '--ignore-default-config')
        param_ignore_default_config='true'
        ;;

      '--ignore-user-config')
        param_ignore_user_config='true'
        ;;

      '--ignore-workspace-config')
        param_ignore_workspace_config='true'
        ;;

      '-vv')
        VERBOSE='true'
        echoinfo_verbose 'Verbose mode enabled'
        echowarn_verbose 'Verbose mode enabled'
        declare -gx MESSAGE_API_SHOW_DEBUG='true'
        echodebug '# -vv mode enabled'
        ;;

      *)
        echofatal "Unknown option: '${param}'"
        usage >&2
        return 1
        ;;
    esac
  done

  case "${param_function}" in
    'all')
      do_shellcheck_for_all "${param_workspace_dir}" || return 100 # Internal error
      ;;

    'sh_only')
      do_shellcheck_for_sh "${param_workspace_dir}" || return 100 # Internal error
      ;;

    *)
      echoerror "Internal error: param_function'${param_function}'"
      usage >&2
      return 1
      ;;
  esac

  if [ "${SHELLCHECK_ERROR}" == 'true' ]; then
    echowarn_verbose "# status=SHELLCHECK_ERROR"
    return 1
  else
    echoinfo_verbose "# status=OK"
  fi
}

main "$@" || exit $?
